///////////////////////////////////////////////////////////////////////////////
// Gen88Font - font generator for PC-8801 emulator
//
// Programed by Manuke 2002

///////////////////////////////////////////////////////////////////////////////
// environment setting

#if defined(_MSC_VER) || defined(__BORLANDC__)

// Compiler is VisualC++ or BorlandC++.

// private operator= are good to have
#pragma warning(disable : 4512)
// identifier was truncated in the debug information
#pragma warning(disable : 4786)

#endif // compiler

///////////////////////////////////////////////////////////////////////////////
// include

#include <stdio.h>
#include <stdlib.h>
#include <mbctype.h>
#include <windows.h>

#include <algorithm>
#include <iterator>
#include <map>

///////////////////////////////////////////////////////////////////////////////
// auto array class template

template <class element_type>
class auto_array {
// typedef
public:
	// my type
	typedef auto_array<element_type> my_type;

// attribute
private:
	// array pointer
	element_type* array_ptr;

// create & destroy
public:
	// normal constructor
	explicit auto_array(element_type* new_array_ptr = NULL) :
		array_ptr(new_array_ptr)
	{
	}
	// copy constructor(not const)
	auto_array(my_type& org) :
		array_ptr(org.release())
	{
	}
	// destructor
	~auto_array() {
		reset();
	}

// operation
public:
	// get array
	element_type* get() const {
		return array_ptr;
	}
	// release & return array pointer
	element_type* release() {
		element_type* result = array_ptr;
		array_ptr = NULL;
		return result;
	}
	// reset array pointer
	void reset(element_type* new_array_ptr = NULL) {
		delete [] array_ptr;
		array_ptr = new_array_ptr;
	}

// operator
public:
	// let
	my_type& operator=(my_type& org) {
		reset(org.release());
		return (*this);
	}
	// ref
	element_type& operator*() const {
		return *get();
	}
	// pointer ref
	element_type* operator->() const {
		return get();
	}
	// indexer
	element_type& operator[](int index) const {
		return *(get()+index);
	}
};

///////////////////////////////////////////////////////////////////////////////
// code converter

class CConv {
// attribute
private:
	// hankaku ascii to zenkaku table
	static int convHankakuAsciiTable_[95];
	// hankaku katakana to zenkaku table
	static int convHankakuKanaTable_[63];
	// zenkaku to hankaku exception table
	static int convZenToHanExceptionTable_[7][2];
	// zenkaku to hankaku map
	static std::map<int, int> convHanToZenMap_;

// operation
public:
	// convert sjis to jis
	static int convSJISToJIS(int sjis) {
		int jis = 0x20,
			hi = (sjis >> 8) & 0xFF,
			lo = sjis & 0xFF;
		if ((hi == 0x00) &&
			((lo <= 0x7E) ||
				((lo >= 0xA0) && (lo <= 0xDF))))
		{
			jis = lo;
		} else if (
			(((hi >= 0x81) && (hi <= 0x9F)) ||
				((hi >= 0xE0) && (hi <= 0xEF))) &&
			(((lo >= 0x40) && (lo <= 0x7E)) ||
				((lo >= 0x80) && (lo <= 0xFC))))
		{
			if (hi >= 0xE0) {
				hi -= 0x40;
			}
			if (lo >= 0x80) {
				lo--;
			}
			int serial = (hi-0x81)*(0x5E*2)+(lo-0x40);
			hi = serial/(0x5E)+0x21;
			lo = (serial%0x5E)+0x21;
			jis = (hi << 8) | lo;
		}
		return jis;
	}
	// convert jis to sjis
	static int convJISToSJIS(int jis) {
		int sjis = 0x20,
			hi = (jis >> 8) & 0xFF,
			lo = jis & 0xFF;
		if ((hi == 0x00) &&
			((lo <= 0x7E) ||
				((lo >= 0xA0) && (lo <= 0xDF))))
		{
			sjis = lo;
		} else if (
			(hi >= 0x21) && (hi <= 0x7E) &&
			(lo >= 0x21) && (lo <= 0x7E))
		{
			int serial = (hi-0x21)*(0x5E)+(lo-0x21);
			hi = serial/(0x5E*2)+0x81;
			lo = (serial%(0x5E*2))+0x40;
			if (hi >= 0xA0) {
				hi += 0x40;
			}
			if (lo >= 0x7F) {
				lo++;
			}
			sjis = (hi << 8) | lo;
		}
		return sjis;
	}
	// convert euc to jis
	static int convEUCToJIS(int euc) {
		unsigned jis = 0x20,
			hi = (euc >> 8) & 0xFF,
			lo = euc & 0xFF;
		if ((hi == 0x00) &&
			(lo <= 0x7E))
		{
			jis = lo;
		} else if (
			(hi == 0x8E) &&
			((lo >= 0xA0) && (lo <= 0xDF)))
		{
			jis = lo;
		} else if (
			(hi >= 0xA1) && (hi <= 0xFE) &&
			(lo >= 0xA1) && (lo <= 0xFE))
		{
			jis = euc & 0x7F7F;
		}
		return jis;
	}
	// convert jis to euc
	static int convJISToEUC(int jis) {
		int euc = 0x20,
			hi = (jis >> 8) & 0xFF,
			lo = jis & 0xFF;
		if ((hi == 0x00) &&
			(lo <= 0x7E))
		{
			euc = lo;
		} else if (
			(hi == 0x00) &&
			((lo >= 0xA0) && (lo <= 0xDF)))
		{
			euc = lo | 0x8E00;
		} else if (
			(hi >= 0x21) && (hi <= 0x7E) &&
			(lo >= 0x21) && (lo <= 0x7E))
		{
			euc = jis | 0x8080;
		}
		return euc;
	}
	// convert jis hankaku to zenkaku
	static int convJISHanToZen(int han) {
		int zen = 0;
		if (han < 0x80) {
			if ((han >= 0x20) && (han <= 0x7E)) {
				zen = convHankakuAsciiTable_[han-0x20];
			}
		} else {
			if ((han >= 0xA1) && (han <= 0xDF)) {
				zen = convHankakuKanaTable_[han-0xA1];
			}
		}
		return zen;
	}
	// convert jis zenkaku to hankaku
	int convJISZenToHan(
		int zen,
		bool* dakutenPtr, bool* handakutenPtr)
	{
		if (convHanToZenMap_.size() <= 0) {
			int i;
			for (i = 0x20; i <= 0x7E; i++) {
				convHanToZenMap_.insert(
					std::make_pair(convHankakuAsciiTable_[i-0x20], i));
			}
			for (i = 0xA1; i <= 0xDF; i++) {
				convHanToZenMap_.insert(
					std::make_pair(convHankakuKanaTable_[i-0xA1], i));
			}
			for (
				i = 0;
				i < sizeof(convZenToHanExceptionTable_)/
					sizeof(convZenToHanExceptionTable_[0]);
				i++)
			{
			convHanToZenMap_.insert(
				std::make_pair(
					convZenToHanExceptionTable_[i][0],
					convZenToHanExceptionTable_[i][1]));
			}
		}
		int han = zen;
		bool dakuten = false, handakuten = false;
		if ((zen >= 0x2421) && (zen <= 0x2473)) {
			// hiragana to katakana
			zen = 0x2500 | (zen & 0xFF);
		}
		if ((zen >= 0x2521) && (zen <= 0x2576)) {
			if ((zen >= 0x252B) && (zen <= 0x2542)) {
				int n = zen-0x252B;
				zen = 0x252B+(n/2)*2;
				switch (n%2) {
				case 1:
					dakuten = true;
					break;
				}
			} else if ((zen >= 0x2544) && (zen <= 0x2549)) {
				int n = zen-0x2544;
				zen = 0x2544+(n/2)*2;
				switch (n%2) {
				case 1:
					dakuten = true;
					break;
				}
			} else if ((zen >= 0x254F) && (zen <= 0x255D)) {
				int n = zen-0x254F;
				zen = 0x254F+(n/3)*3;
				switch (n%3) {
				case 1:
					dakuten = true;
					break;
				case 2:
					handakuten = true;
					break;
				}
			} else if (zen == 0x2574) {
				zen = 0x2526;
				dakuten = true;
			}
		}
		std::map<int, int>::const_iterator hankakuIt =
			convHanToZenMap_.find(zen);
		if (hankakuIt != convHanToZenMap_.end()) {
			han = (*hankakuIt).second;
		}
		if (dakutenPtr != NULL) {
			*dakutenPtr = dakuten;
		}
		if (handakutenPtr != NULL) {
			*handakutenPtr = handakuten;
		}
		return han;
	}
};

int CConv::convHankakuAsciiTable_[] = {
	0x2121, 0x212A, 0x2149, 0x2174, 0x2170, 0x2173, 0x2175, 0x2147,
	0x214A, 0x214B, 0x2176, 0x215C, 0x2124, 0x215D, 0x2125, 0x213F,
	0x2330, 0x2331, 0x2332, 0x2333, 0x2334, 0x2335, 0x2336, 0x2337,
	0x2338, 0x2339, 0x2127, 0x2128, 0x2163, 0x2161, 0x2164, 0x2129,
	0x2177, 0x2341, 0x2342, 0x2343, 0x2344, 0x2345, 0x2346, 0x2347,
	0x2348, 0x2349, 0x234A, 0x234B, 0x234C, 0x234D, 0x234E, 0x234F,
	0x2350, 0x2351, 0x2352, 0x2353, 0x2354, 0x2355, 0x2356, 0x2357,
	0x2358, 0x2359, 0x235A, 0x214E, 0x216F, 0x214F, 0x2130, 0x2132,
	0x2146, 0x2361, 0x2362, 0x2363, 0x2364, 0x2365, 0x2366, 0x2367,
	0x2368, 0x2369, 0x236A, 0x236B, 0x236C, 0x236D, 0x236E, 0x236F,
	0x2370, 0x2371, 0x2372, 0x2373, 0x2374, 0x2375, 0x2376, 0x2377,
	0x2378, 0x2379, 0x237A, 0x2150, 0x2143, 0x2151, 0x2131
};

int CConv::convHankakuKanaTable_[] = {
	        0x2123, 0x2156, 0x2157, 0x2122, 0x2126, 0x2572, 0x2521,
	0x2523, 0x2525, 0x2527, 0x2529, 0x2563, 0x2565, 0x2567, 0x2543,
	0x213C, 0x2522, 0x2524, 0x2526, 0x2528, 0x252A, 0x252B, 0x252D,
	0x252F, 0x2531, 0x2533, 0x2535, 0x2537, 0x2539, 0x253B, 0x253D,
	0x253F, 0x2541, 0x2544, 0x2546, 0x2548, 0x254A, 0x254B, 0x254C,
	0x254D, 0x254E, 0x254F, 0x2552, 0x2555, 0x2558, 0x255B, 0x255E,
	0x255F, 0x2560, 0x2561, 0x2562, 0x2564, 0x2566, 0x2568, 0x2569,
	0x256A, 0x256B, 0x256C, 0x256D, 0x256F, 0x2573, 0x212B, 0x212C
};

int CConv::convZenToHanExceptionTable_[][2] = {
	{ 0x212D, 0x60 },
	{ 0x2141, 0x7E },
	{ 0x256E, 0xDC },
	{ 0x2570, 0xB2 },
	{ 0x2571, 0xB4 },
	{ 0x2575, 0xB6 },
	{ 0x2576, 0xB9 }
};

std::map<int, int> CConv::convHanToZenMap_;

///////////////////////////////////////////////////////////////////////////////
// image class

class Image {
// attribute
private:
	// bitmap width
	int width_;
	// bitmap height
	int height_;
	// bitmap pixel data
	auto_array<int> pixels_;

// create & destroy
public:
	// constructor
	Image() :
		width_(0), height_(0)
	{
	}
	// destructor
	~Image() {
	}
	// create new image
	void create(int width, int height) {
		width_ = width;
		height_ = height;
		int length = getWidth()*getHeight();
		pixels_.reset((length > 0)? new int[length]: NULL);
	}

// operation
public:
	// get width
	int getWidth() const {
		return width_;
	}
	// get height
	int getHeight() const {
		return height_;
	}
	// get pixel
	int getPixel(int x, int y) const {
		int pixel = -1;
		if ((x >= 0) && (x < getWidth()) &&
			(y >= 0) && (y < getHeight()))
		{
			pixel = pixels_[x+y*getWidth()];
		}
		return pixel;
	}
	// set pixel
	void setPixel(int x, int y, int pixel) {
		if ((x >= 0) && (x < getWidth()) &&
			(y >= 0) && (y < getHeight()))
		{
			pixels_[x+y*getWidth()] = pixel;
		}
	}
};

///////////////////////////////////////////////////////////////////////////////
// bitmap loader

class BitmapImageLoader {
// create & destroy
public:
	// constructor
	BitmapImageLoader() {
	}
	// destructor
	~BitmapImageLoader() {
	}

// utility
private:
	// read uint8 data from file
	static unsigned readUINT8(FILE* fpt) {
		unsigned char buf[1];
		if (fread(buf, 1, sizeof(buf), fpt) != sizeof(buf)) {
			throw "file read error";
		}
		return buf[0];
	}
	// read uint16 data from file
	static unsigned readUINT16(FILE* fpt) {
		unsigned char buf[2];
		if (fread(buf, 1, sizeof(buf), fpt) != sizeof(buf)) {
			throw "file read error";
		}
		return buf[0]+(buf[1] << 8);
	}
	// read uint32 data from file
	static unsigned readUINT32(FILE* fpt) {
		unsigned char buf[4];
		if (fread(buf, 1, sizeof(buf), fpt) != sizeof(buf)) {
			throw "file read error";
		}
		return buf[0]+(buf[1] << 8)+(buf[2] << 16)+(buf[3] << 24);
	}

// operation
public:
// load image
	void loadImage(Image& image, const char* filename) {
		FILE* fpt = NULL;
		try {
			fpt = fopen(filename, "rb");
			if (fpt == NULL) {
				throw "cannot open bitmap";
			}
			// bitmap file header
			struct BitmapFileHeader {
				int bfType,
					bfSize,
					bfReserved1,
					bfReserved2,
					bfOffBits;
			} bmfh;
			bmfh.bfType = readUINT16(fpt);
			bmfh.bfSize = readUINT32(fpt);
			bmfh.bfReserved1 = readUINT16(fpt);
			bmfh.bfReserved2 = readUINT16(fpt);
			bmfh.bfOffBits = readUINT32(fpt);
			if (bmfh.bfType != ('B'+('M' << 8))) {
				throw "bad format bitmap";
			}
			// bitmap info header
			struct BitmapInfoHeader {
				int biSize,
					biWidth,
					biHeight,
					biPlanes,
					biBitCount,
					biCompression,
					biSizeImage,
					biXPelsPerMeter,
					biYPelsPerMeter,
					biClrUsed,
					biClrImportant;
			} bmih;
			bmih.biSize = readUINT32(fpt);
			bmih.biWidth = readUINT32(fpt);
			bmih.biHeight = readUINT32(fpt);
			bmih.biPlanes = readUINT16(fpt);
			bmih.biBitCount = readUINT16(fpt);
			bmih.biCompression = readUINT32(fpt);
			bmih.biSizeImage = readUINT32(fpt);
			bmih.biXPelsPerMeter = readUINT32(fpt);
			bmih.biYPelsPerMeter = readUINT32(fpt);
			bmih.biClrUsed = readUINT32(fpt);
			bmih.biClrImportant = readUINT32(fpt);
			const int stdInfoHeaderSize = 4+4+4+2+2+4+4+4+4+4+4;
			if (bmih.biSize < stdInfoHeaderSize) {
				throw "bad format bitmap";
			}
			bool reverse = (bmih.biHeight < 0);
			if (reverse) {
				bmih.biHeight = -bmih.biHeight;
			}
			if ((bmih.biWidth < 1) || (bmih.biWidth > 4096) ||
				(bmih.biHeight < 1) || (bmih.biHeight > 4096) ||
				(bmih.biPlanes != 1) ||
				((bmih.biBitCount != 1) && (bmih.biBitCount != 2) &&
					(bmih.biBitCount != 4) && (bmih.biBitCount != 8) &&
					(bmih.biBitCount != 16) && (bmih.biBitCount != 24) &&
					(bmih.biBitCount != 32)) ||
				((bmih.biCompression != 0) && (bmih.biCompression != 3)))
			{
				throw "bad format bitmap";
			}
			if (bmih.biSize > stdInfoHeaderSize) {
				int skip = bmih.biSize-stdInfoHeaderSize;
				if (fseek(fpt, skip, SEEK_CUR) != 0) {
					throw "file read error";
				}
			}
			// bitfield
			int ofsR = 0, ofsG = 0, ofsB = 0,
				bitsR = 1, bitsG = 1, bitsB = 1;
			if (bmih.biCompression != 3) {
				if (bmih.biBitCount == 16) {
					ofsR = 10;
					ofsG = 5;
					ofsB = 0;
					bitsR = bitsG = bitsB = 5;
				} else {
					ofsR = 16;
					ofsG = 8;
					ofsB = 0;
					bitsR = bitsG = bitsB = 8;
				}
			} else {
				for (int i = 0; i < 3; i++) {
					int mask = readUINT32(fpt);
					int ofs, bits;
					for (ofs = 0; ofs < 32; ofs++) {
						if ((mask & 1) != 0) {
							break;
						}
						mask >>= 1;
					}
					for (bits = 0; bits+ofs < 32; bits++) {
						if ((mask & 1) == 0) {
							break;
						}
						mask >>= 1;
					}
					switch (i) {
					case 0:
						ofsR = ofs;
						bitsR = bits;
						break;
					case 1:
						ofsG = ofs;
						bitsG = bits;
						break;
					case 2:
						ofsB = ofs;
						bitsB = bits;
						break;
					}
				}
			}
			// palette
			int palCount = (bmih.biClrUsed >= bmih.biClrImportant)?
					bmih.biClrUsed:
					bmih.biClrImportant;
			bool indexColor = (bmih.biBitCount <= 8);
			if ((palCount == 0) && indexColor) {
				palCount = 1 << bmih.biBitCount;
			}
			auto_array<int> palette;
			if (palCount > 0) {
				palette.reset(new int[palCount]);
				for (int i = 0; i < palCount; i++) {
					palette[i] = readUINT32(fpt) & 0x00FFFFFF;
				}
			}
			// create image
			image.create(bmih.biWidth, bmih.biHeight);
			// bit data
			unsigned scanline = ((bmih.biWidth*bmih.biBitCount+31)/32)*4;
			auto_array<unsigned char> buf(new unsigned char[scanline]);
			for (int i = 0; i < bmih.biHeight; i++) {
				if (fread(buf.get(), 1, scanline, fpt) != scanline) {
					throw "file read error";
				}
				int y = reverse? i: (bmih.biHeight-i-1);
				int ofs = 0;
				for (int x = 0; x < bmih.biWidth; x++) {
					int pixel = 0;
					if (!indexColor) {
						int data;
						if (bmih.biBitCount == 32) {
							data = buf[ofs]+
								(buf[ofs+1] << 8)+
								(buf[ofs+2] << 16)+
								(buf[ofs+3] << 24);
							ofs += 4;
						} else if (bmih.biBitCount == 24) {
							data = buf[ofs]+
								(buf[ofs+1] << 8)+
								(buf[ofs+2] << 16);
							ofs += 3;
						} else {
							data = buf[ofs]+
								(buf[ofs+1] << 8);
							ofs += 2;
						}
						int red = (data >> ofsR) & ((1 << bitsR)-1);
						if (bitsR < 8) {
							int left = (red << (8-bitsR));
							red = 0;
							while (left > 0) {
								red |= left;
								left >>= bitsR;
							}
						} else if (bitsR > 8) {
							red >>= bitsR-8;
						}
						int green = (data >> ofsG) & ((1 << bitsG)-1);
						if (bitsG < 8) {
							int left = (green << (8-bitsG));
							green = 0;
							while (left > 0) {
								green |= left;
								left >>= bitsG;
							}
						} else if (bitsG > 8) {
							green >>= bitsG-8;
						}
						int blue = (data >> ofsB) & ((1 << bitsB)-1);
						if (bitsB < 8) {
							int left = (blue << (8-bitsB));
							blue = 0;
							while (left > 0) {
								blue |= left;
								left >>= bitsB;
							}
						} else if (bitsB > 8) {
							blue >>= bitsB-8;
						}
						pixel = ((red & 0xFF) << 16)+
							((green & 0xFF) << 8)+
							(blue & 0xFF);
					} else {
						ofs = x*bmih.biBitCount;
						int pal =
							((buf[ofs/8] << (ofs%8)) & 0xFF) >>
								(8-bmih.biBitCount);
						if (pal < palCount) {
							pixel = palette[pal];
						}
					}
					image.setPixel(x, y, pixel);
				}
			}
			fclose(fpt);
			fpt = NULL;
		} catch (...) {
			if (fpt != NULL) {
				fclose(fpt);
			}
			throw;
		}
	}
};

///////////////////////////////////////////////////////////////////////////////
// gui font drawer

class GUIFontDrawer {
// enum
public:
	enum {
		FONT_WIDTH = 16,
		FONT_HEIGHT = 16
	};

// attribute
private:
	// font name
	char* fontName_;
	// draw device context
	HDC drawDC_;
	// draw bitmap
	HBITMAP drawBitmap_;
	// original bitmap
	HBITMAP orgBitmap_;
	// draw font
	HFONT drawFont_;
	// original font
	HFONT orgFont_;

// create & destroy
public:
	GUIFontDrawer() :
		fontName_(NULL),
		drawDC_(NULL),
		drawBitmap_(NULL), orgBitmap_(NULL),
		drawFont_(NULL), orgFont_(NULL)
	{
	}
	~GUIFontDrawer() {
		dispose();
	}
	// initialize
	void initialize(const char* fontName) {
		dispose();
		HDC desktopDC = NULL;
		try {
			int length = strlen(fontName);
			if ((length > 2) &&
				((fontName[0] == '"') && (fontName[length] == '"')) ||
				((fontName[0] == '\'') && (fontName[length] == '\'')))
			{
				fontName++;
				length -= 2;
			}
			fontName_ = new char[length+1];
			strncpy(fontName_, fontName, length);
			fontName_[length] = '\0';
			drawBitmap_ = CreateBitmap(
				FONT_WIDTH, FONT_HEIGHT, 1, 1, NULL);
			if (drawBitmap_ == NULL) {
				throw "GDI error";
			}
			desktopDC = GetDC(HWND_DESKTOP);
			if (desktopDC == NULL) {
				throw "GDI error";
			}
			drawDC_ = CreateCompatibleDC(desktopDC);
			if (drawDC_ == NULL) {
				throw "GDI error";
			}
			orgBitmap_ = (HBITMAP)SelectObject(drawDC_, drawBitmap_);
			ReleaseDC(HWND_DESKTOP, desktopDC);
			desktopDC = NULL;
			LOGFONT lf;
			memset(&lf, 0, sizeof(lf));
			lf.lfHeight = FONT_HEIGHT;
			lf.lfWeight = FW_NORMAL;
			lf.lfPitchAndFamily = FIXED_PITCH;
			lf.lfCharSet = SHIFTJIS_CHARSET;
			strcpy(lf.lfFaceName, fontName_);
			drawFont_ = CreateFontIndirect(&lf);
			if (drawFont_ == NULL) {
				throw "cannot create font";
			}
			orgFont_ = (HFONT)SelectObject(drawDC_, drawFont_);
		} catch (...) {
			if (desktopDC != NULL) {
				ReleaseDC(HWND_DESKTOP, desktopDC);
			}
			dispose();
			throw;
		}
		if (desktopDC != NULL) {
			ReleaseDC(HWND_DESKTOP, desktopDC);
		}
	}
	// dispose
	void dispose() {
		if (drawDC_ != NULL) {
			if (orgBitmap_ != NULL) {
				SelectObject(drawDC_, orgBitmap_);
				orgBitmap_ = NULL;
			}
			if (orgFont_ != NULL) {
				SelectObject(drawDC_, orgFont_);
				orgFont_ = NULL;
			}
			DeleteDC(drawDC_);
			drawDC_ = NULL;
		}
		if (drawBitmap_ != NULL) {
			DeleteObject(drawBitmap_);
			drawBitmap_ = NULL;
		}
		if (drawFont_ != NULL) {
			DeleteObject(drawFont_);
			drawFont_ = NULL;
		}
		delete [] fontName_;
		fontName_ = NULL;
	}

// operation
public:
	// drawImage
	void drawImage(Image& image, int jiscode) {
		if ((image.getWidth() != FONT_WIDTH) ||
			(image.getHeight() != FONT_HEIGHT))
		{
			image.create(FONT_WIDTH, FONT_HEIGHT);
		}
		RECT drawRect;
		SetRect(&drawRect, 0, 0, FONT_WIDTH, FONT_HEIGHT);
		char str[3];
		int length = 0;
		int hi = (jiscode >> 8) & 0xFF,
			lo = jiscode & 0xFF;
		if (hi == 0x00) {
			if (lo <= 0x1F) {
			} else if (
				((lo >= 0x80) && (lo <= 0x9F)) ||
				(lo >= 0xE0))
			{
			} else {
				str[length++] = (char)lo;
			}
		} else {
			if (((hi < 0x21) || (hi > 0x7E)) ||
				((lo < 0x21) || (lo > 0x7E)) ||
				((jiscode >= 0x222F) && (jiscode <= 0x227E)) ||
				((jiscode >= 0x2321) && (jiscode <= 0x232F)) ||
				((jiscode >= 0x233A) && (jiscode <= 0x2340)) ||
				((jiscode >= 0x235B) && (jiscode <= 0x2360)) ||
				((jiscode >= 0x237B) && (jiscode <= 0x237E)) ||
				((jiscode >= 0x2474) && (jiscode <= 0x247E)) ||
				((jiscode >= 0x2577) && (jiscode <= 0x257E)) ||
				((jiscode >= 0x2639) && (jiscode <= 0x2640)) ||
				((jiscode >= 0x2659) && (jiscode <= 0x267F)) ||
				((jiscode >= 0x2742) && (jiscode <= 0x2750)) ||
				((jiscode >= 0x2772) && (jiscode <= 0x277E)))
			{
			} else {
				int sjiscode = CConv::convJISToSJIS(jiscode);
				str[length++] = (char)((sjiscode >> 8) & 0xFF);
				str[length++] = (char)(sjiscode & 0xFF);
			}
		}
		str[length] = '\0';
		ExtTextOut(
			drawDC_, 0, 0, ETO_OPAQUE, &drawRect, str, length, NULL);
		for (int y = 0; y < FONT_HEIGHT; y++) {
			for (int x = 0; x < FONT_WIDTH; x++) {
				image.setPixel(
					x, y,
					GetPixel(drawDC_, x, y) & 0x00FFFFFF);
			}
		}
	}
};

///////////////////////////////////////////////////////////////////////////////
// PC-8801 font builder

class PC88FontBuilder {
// enum
public:
	// rom type
	enum {
		ROM_KANJI1,
		ROM_KANJI2,
		ROM_PC88,
		ROM_HIRAFONT
	};

// attribute
private:
	// kanji rom1 data
	unsigned char kanjiRom1_[131072];
	// kanji rom2 data
	unsigned char kanjiRom2_[131072];
	// hiragana rom data
	unsigned char hiranagaRom_[512];

// create & destroy
public:
	// constructor
	PC88FontBuilder() {
		memset(kanjiRom1_, 0, sizeof(kanjiRom1_));
		memset(kanjiRom2_, 0, sizeof(kanjiRom2_));
		memset(hiranagaRom_, 0, sizeof(hiranagaRom_));
	}
	// destructor
	~PC88FontBuilder() {
	}

// operation
public:
	// get address
	static int getAddress(int jiscode) {
		int address = -1;
		int hi = (jiscode >> 8) & 0xFF,
			lo = jiscode & 0xFF;
		if (hi == 0x00) {
			address = 0x0000 |
				((jiscode & 0x00FF) << 3);
		} else if (hi == 0x01) {
			address = 0x0800 |
				((jiscode & 0x00FF) << 2);
		} else if ((lo >= 0x21) && (lo <= 0x7E)) {
			if (((jiscode >= 0x2121) && (jiscode <= 0x277E)) ||
				((jiscode >= 0x7021) && (jiscode <= 0x737E)))
			{
				address = 0x0000 |
					((jiscode & 0x001F) << 4) |
					((jiscode & 0x0060) << 7) |
					((jiscode & 0x0700) << 1);
			} else if (
				((jiscode >= 0x3021) && (jiscode <= 0x4F53)) ||
				((jiscode >= 0x5021) && (jiscode <= 0x6F7E)))
			{
				address = 0x0000 |
					((jiscode & 0x001F) << 4) |
					((jiscode & 0x0060) << 9) |
					((jiscode & 0x0F00) << 1) |
					(jiscode & 0x2000);
			}
		}
		return address;
	}
	// get kanji rom1 data
	unsigned char* getKanjiRom1() {
		return kanjiRom1_;
	}
	// get kanji rom2 data
	unsigned char* getKanjiRom2() {
		return kanjiRom2_;
	}
	// get hiragana rom data
	unsigned char* getHiranagaRom() {
		return hiranagaRom_;
	}
	// build small font
	void buildSmallFont() {
		Image image;
		BitmapImageLoader loader;
		loader.loadImage(image, "SmallFont.bmp");
		int address = getAddress(0x0100)*2;
		int i;
		for (i = 0; i < 256; i++) {
			for (int y = 0; y < 8; y++) {
				int data = 0;
				for (int x = 0; x < 8; x++) {
					if (image.getPixel(
							(i/16)*9+1+x,
							(i%16)*9+1+y) == 0)
					{
						data |= 0x80 >> x;
					}
				}
				kanjiRom1_[address++] = (unsigned char)data;
			}
		}
		for (i = 0; i < 256+128; i++) {
			for (int y = 0; y < 8; y++) {
				kanjiRom1_[address++] = 0x40;
			}
		}
	}
	// build half font
	void buildHalfFont() {
		Image image;
		BitmapImageLoader loader;
		loader.loadImage(image, "HalfFont.bmp");
		int address = getAddress(0x0000)*2;
		for (int i = 0; i < 256; i++) {
			for (int y = 0; y < 16; y++) {
				int data = 0;
				for (int x = 0; x < 8; x++) {
					if (image.getPixel(
							(i/16)*9+1+x,
							(i%16)*17+1+y) == 0)
					{
						data |= 0x80 >> x;
					}
				}
				kanjiRom1_[address++] = (unsigned char)data;
			}
		}
	}
	// build full font
	void buildFullFont(const char* fontName) {
		Image image;
		GUIFontDrawer drawer;
		drawer.initialize(fontName);
		for (int hi = 0x21; hi <= 0x7E; hi++) {
			for (int lo = 0x21; lo <= 0x7E; lo++) {
				int jiscode = (hi << 8) | lo;
				int address = getAddress(jiscode);
				if (address >= 0) {
					address *= 2;
					unsigned char* rom = (jiscode < 0x5000)?
						kanjiRom1_:
						kanjiRom2_;
					drawer.drawImage(image, jiscode);
					for (int y = 0; y < 16; y++) {
						for (int x1 = 0; x1 < 16; x1 += 8) {
							int data = 0;
							for (int x2 = 0; x2 < 8; x2++) {
								if (image.getPixel(
										x1+x2,
										y) == 0)
								{
									data |= 0x80 >> x2;
								}
							}
							rom[address++] = (unsigned char)data;
						}
					}
				}
			}
		}
	}
	// build hiragana font
	void buildHiraganaFont() {
		Image image;
		BitmapImageLoader loader;
		loader.loadImage(image, "HiraganaFont.bmp");
		int address = 0;
		for (int i = 0; i < 64; i++) {
			for (int y = 0; y < 8; y++) {
				int data = 0;
				for (int x = 0; x < 8; x++) {
					if (image.getPixel(
							(i/16)*9+1+x,
							(i%16)*9+1+y) == 0)
					{
						data |= 0x80 >> x;
					}
				}
				hiranagaRom_[address++] = (unsigned char)data;
			}
		}
	}
	// create rom file
	void createFile(int romType) {
		FILE* fpt = NULL;
		try {
			const char* fileName;
			switch (romType) {
			case ROM_KANJI1:
				fileName = "KANJI1.ROM";
				break;
			case ROM_KANJI2:
				fileName = "KANJI2.ROM";
				break;
			case ROM_PC88:
				fileName = "PC88.FNT";
				break;
			case ROM_HIRAFONT:
				fileName = "HIRAFONT.ROM";
				break;
			default:
				throw "bad rom type";
			}
			fpt = fopen(fileName, "wb");
			if (fpt == NULL) {
				throw "cannot create file";
			}
			switch (romType) {
			case ROM_KANJI1:
				if (fwrite(kanjiRom1_, 1, sizeof(kanjiRom1_), fpt) !=
						sizeof(kanjiRom1_))
				{
					throw "file write error";
				}
				break;
			case ROM_KANJI2:
				if (fwrite(kanjiRom2_, 1, sizeof(kanjiRom2_), fpt) !=
						sizeof(kanjiRom2_))
				{
					throw "file write error";
				}
				break;
			case ROM_PC88:
				if ((fwrite(kanjiRom1_+0x1000, 1, 0x0800, fpt) != 0x0800) ||
					(fwrite(kanjiRom1_, 1, 0x1000, fpt) != 0x1000))
				{
					throw "file write error";
				}
				break;
			case ROM_HIRAFONT:
				if (fwrite(hiranagaRom_, 1, sizeof(hiranagaRom_), fpt) !=
						sizeof(hiranagaRom_))
				{
					throw "file write error";
				}
				break;
			}
			if (fflush(fpt) != 0) {
				throw "cannot flush file";
			}
			if (fclose(fpt) != 0) {
				throw "cannot close file";
			}
			fpt = NULL;
		} catch (...) {
			if (fpt != NULL) {
				fclose(fpt);
			}
			throw;
		}
	}
};

///////////////////////////////////////////////////////////////////////////////
// debug window

#ifdef _DEBUG

class DebugWindow {
// enum
private:
	// debug size
	enum {
		DEBUG_ZOOM = 2,
		DEBUG_WIDTH = (8*DEBUG_ZOOM+1)*33+1,
		DEBUG_HEIGHT = (8*DEBUG_ZOOM+1)*32+1,
	};

// attribute
private:
	// show page
	int showPage_;
	// temporary bitmap
	HBITMAP tmpBitmap_;
	// builder
	PC88FontBuilder builder_;
	// PC-8801 KANJI1.ROM
	unsigned char pc88KanjiRom1_[131072];
	// PC-8801 KANJI2.ROM
	unsigned char pc88KanjiRom2_[131072];
	// ELISA HIRAFONT.ROM
	unsigned char elisaHiraganaRom_[512];
	// window mapper
	static std::map<HWND, DebugWindow*> mapper;

// create & destroy
public:
	// constructor
	DebugWindow(const char* fontName) :
		showPage_(-1),
		tmpBitmap_(NULL)
	{
		builder_.buildSmallFont();
		builder_.buildHalfFont();
		builder_.buildFullFont(fontName);
		builder_.buildHiraganaFont();
		FILE* fpt;
		memset(pc88KanjiRom1_, 0, sizeof(pc88KanjiRom1_));
		fpt = fopen("PC88\\KANJI1.ROM", "rb");
		if (fpt != NULL) {
			fread(pc88KanjiRom1_, sizeof(pc88KanjiRom1_), 1, fpt);
			fclose(fpt);
		}
		memset(pc88KanjiRom2_, 0, sizeof(pc88KanjiRom2_));
		fpt = fopen("PC88\\KANJI2.ROM", "rb");
		if (fpt != NULL) {
			fread(pc88KanjiRom2_, sizeof(pc88KanjiRom2_), 1, fpt);
			fclose(fpt);
		}
		memset(elisaHiraganaRom_, 0, sizeof(elisaHiraganaRom_));
		fpt = fopen("ELISA\\HIRAFONT.ROM", "rb");
		if (fpt != NULL) {
			fread(elisaHiraganaRom_, sizeof(elisaHiraganaRom_), 1, fpt);
			fclose(fpt);
		}
	}
	// destructor
	~DebugWindow() {
	}

// window procedure
private:
	// debug window procudure
	static LRESULT CALLBACK debugWndProc(
		HWND wnd, UINT message, WPARAM wParam, LPARAM lParam)
	{
		LRESULT result = 0;
		if (message == WM_NCCREATE) {
			mapper[wnd] = (DebugWindow*)(
				((CREATESTRUCT*)lParam)->lpCreateParams);
		}
		std::map<HWND, DebugWindow*>::iterator it = mapper.find(wnd);
		if (it != mapper.end()) {
			result = ((*it).second)->debugWndProcX(
				wnd, message, wParam, lParam);
		}
		if (message == WM_NCDESTROY) {
			mapper.erase(wnd);
		}
		return result;
	}
	// debug window procudure(not static)
	LRESULT debugWndProcX(
		HWND wnd, UINT message, WPARAM wParam, LPARAM lParam)
	{
		bool dwp = false;
		LRESULT result = 0;
		switch (message) {
		case WM_CREATE:
			break;
		case WM_DESTROY:
			if (tmpBitmap_ != NULL) {
				DeleteObject(tmpBitmap_);
				tmpBitmap_ = NULL;
			}
			PostQuitMessage(0);
			break;
		case WM_LBUTTONDOWN:
			if (showPage_ < 63) {
				showPage_++;
				if (tmpBitmap_ != NULL) {
					DeleteObject(tmpBitmap_);
					tmpBitmap_ = NULL;
				}
				InvalidateRect(wnd, NULL, TRUE);
			}
			break;
		case WM_RBUTTONDOWN:
			if (showPage_ > -1) {
				showPage_--;
				if (tmpBitmap_ != NULL) {
					DeleteObject(tmpBitmap_);
					tmpBitmap_ = NULL;
				}
				InvalidateRect(wnd, NULL, TRUE);
			}
			break;
		case WM_PAINT:
			{ // dummy block
				PAINTSTRUCT ps;
				HDC winDC = BeginPaint(wnd, &ps);
				bool redraw = false;
				if (tmpBitmap_ == NULL) {
					redraw = true;
					tmpBitmap_ = CreateCompatibleBitmap(
						winDC, DEBUG_WIDTH, DEBUG_HEIGHT);
				}
				if (tmpBitmap_ != NULL) {
					HDC memDC = CreateCompatibleDC(winDC);
					if (memDC != NULL) {
						HBITMAP orgBitmap =
							(HBITMAP)SelectObject(memDC, tmpBitmap_);
						if (redraw) {
							buildScreen(memDC);
						}
						SetStretchBltMode(winDC, STRETCH_DELETESCANS);
						BitBlt(
							winDC,
							0, 0, DEBUG_WIDTH, DEBUG_HEIGHT,
							memDC,
							0, 0,
							SRCCOPY);
						SelectObject(memDC, orgBitmap);
						DeleteDC(memDC);
					}
				}
				EndPaint(wnd, &ps);
			}
			break;
		default:
			dwp = true;
			break;
		}
		return dwp? DefWindowProc(wnd, message, wParam, lParam) : result;
	}

// operation
private:
	// build screen
	void buildScreen(HDC dc) {
		int cx, cy, bytes = 4096;
		if (showPage_ < 0) {
			cx = 8;
			cy = 8;
			bytes = 512;
		} else if (showPage_ == 0) {
			cx = 8;
			cy = 16;
		} else if (showPage_ == 1) {
			cx = 8;
			cy = 8;
		} else {
			cx = 16;
			cy = 16;
		}
		int bx = 0, by = 0;
		HPEN greenPen = CreatePen(PS_SOLID, 1, RGB(128, 255, 128));
		HPEN orgPen = (HPEN)SelectObject(dc, greenPen);
		HBRUSH orgBrush = (HBRUSH)SelectObject(dc, GetStockObject(WHITE_BRUSH));
		RECT rectDraw;
		SetRect(
			&rectDraw,
			0, 0, DEBUG_WIDTH, DEBUG_HEIGHT);
		ExtTextOut(
			dc, 0, 0,
			ETO_OPAQUE, &rectDraw,
			"", 0, NULL);
		SetBkColor(dc, RGB(0, 0, 0));
		for (int i = 0; i < 2; i++) {
			unsigned char* rom;
			if (showPage_ < 0) {
				rom = (i == 0)?
					builder_.getHiranagaRom():
					elisaHiraganaRom_;
			} else {
				if (showPage_ < 32) {
					rom = (i == 0)?
						builder_.getKanjiRom1():
						pc88KanjiRom1_;
				} else {
					rom = (i == 0)?
						builder_.getKanjiRom2():
						pc88KanjiRom2_;
				}
				rom += (showPage_%32)*4096;
			}
			for (int ofs = 0; ofs < bytes;) {
				Rectangle(
					dc,
					bx, by,
					bx+cx*DEBUG_ZOOM+2,
					by+cy*DEBUG_ZOOM+2);
				for (int y = 0; y < cy; y++) {
					for (int x1 = 0; x1 < cx; x1 += 8) {
						int data = rom[ofs++];
						for (int x2 = 0; x2 < 8; x2++) {
							if ((data & (0x80 >> x2)) != 0) {
								SetRect(
									&rectDraw,
									bx+1+(x1+x2)*DEBUG_ZOOM,
									by+1+y*DEBUG_ZOOM,
									bx+1+(x1+x2+1)*DEBUG_ZOOM,
									by+1+(y+1)*DEBUG_ZOOM);
								ExtTextOut(
									dc,
									0, 0, ETO_OPAQUE, &rectDraw,
									"", 0, NULL);
							}
						}
					}
				}
				by += cy*DEBUG_ZOOM+1;
				if (by+cy*DEBUG_ZOOM+1 >= DEBUG_HEIGHT) {
					by = 0;
					bx += cx*DEBUG_ZOOM+1;
				}
			}
			bx += cx*DEBUG_ZOOM+1;
		}
		SelectObject(dc, orgPen);
		SelectObject(dc, orgBrush);
		DeleteObject(greenPen);
	}

public:
	// show window
	void show() {
		WNDCLASS wc;
		memset(&wc, 0, sizeof(wc));
		wc.lpfnWndProc = debugWndProc;
		wc.hInstance = (HINSTANCE)GetModuleHandle(NULL);
		wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
		wc.hCursor = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
		wc.lpszClassName = "debug window";
		if (RegisterClass(&wc) != NULL) {
			RECT debugRect;
			SetRect(&debugRect, 0, 0, DEBUG_WIDTH, DEBUG_HEIGHT);
			AdjustWindowRect(&debugRect, WS_OVERLAPPEDWINDOW, FALSE);
			HWND wnd = CreateWindow(
				"debug window", "Gen88Font : debug",
				WS_OVERLAPPEDWINDOW | WS_VISIBLE,
				0,
				0,
				debugRect.right-debugRect.left,
				debugRect.bottom-debugRect.top,
				NULL, NULL, wc.hInstance, this);
			if (wnd != NULL) {
				ShowWindow(wnd, SW_NORMAL);
				MSG msg;
				while (GetMessage(&msg, NULL, 0, 0)) {
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}
		}
	}
};

// window mapper

std::map<HWND, DebugWindow*> DebugWindow::mapper;

#endif // _DEBUG

///////////////////////////////////////////////////////////////////////////////
// main

// main routine

int main(int argc, char* argv[]) {
	try {
		bool romM88 = false,
			romP88SR = false,
			romHiragana = false;
		const char* defFontName = "lr SVbN";
		const char* fontName = defFontName;
		for (int i = 1; i < argc; i++) {
			if (stricmp(argv[i], "--help") == 0) {
				romM88 = romP88SR = romHiragana = false;
			} else if (stricmp(argv[i], "--m88") == 0) {
				romM88 = true;
			} else if (stricmp(argv[i], "--p88sr") == 0) {
				romP88SR = true;
			} else if (stricmp(argv[i], "--hiragana") == 0) {
				romHiragana = true;
			} else if ((stricmp(argv[i], "--font") == 0) && (i < argc-1)) {
				fontName = argv[++i];
			}
		}

#ifdef _DEBUG

		DebugWindow window(fontName);
		window.show();

#endif // _DEBUG

		if (!romM88 && !romP88SR && !romHiragana) {
			printf(
				"Gen88Font - font generator for PC-8801 emulator\n"
				"\n"
				"usage:\n"
				"\t\tGen88Font [options...]\n"
				"\n"
				"option:\n"
				"\t--help\n"
				"\t\tshow help(this message)\n"
				"\t--m88\n"
				"\t\tgenerate M88 font(KANJI1.ROM, KANJI2.ROM)\n"
				"\t--p88sr\n"
				"\t\tgenerate P88SR font(PC88.FNT)\n"
				"\t--hiragana\n"
				"\t\tgenerate hiragana font(HIRAFONT.ROM)\n"
				"\t--font <font name>\n"
				"\t\tchange full-width font name\n"
				"\t\t(default : \"%s\")\n",
				defFontName);
		} else {
			PC88FontBuilder builder;
			if (romM88 || romP88SR) {
				printf("building quarter-size font...");
				builder.buildSmallFont();
				printf("done.\n");
				printf("building half-width font...");
				builder.buildHalfFont();
				printf("done.\n");
			}
			if (romM88) {
				printf("building full-width font(used \"%s\")...", fontName);
				builder.buildFullFont(fontName);
				printf("done.\n");
			}
			if (romHiragana) {
				printf("building hiragana font...");
				builder.buildHiraganaFont();
				printf("done.\n");
			}
			if (romM88) {
				printf("creating 'KANJI1.ROM'...");
				builder.createFile(PC88FontBuilder::ROM_KANJI1);
				printf("done.\n");
				printf("creating 'KANJI2.ROM'...");
				builder.createFile(PC88FontBuilder::ROM_KANJI2);
				printf("done.\n");
			}
			if (romP88SR) {
				printf("creating 'PC88.FNT'...");
				builder.createFile(PC88FontBuilder::ROM_PC88);
				printf("done.\n");
			}
			if (romHiragana) {
				printf("creating 'HIRAFONT.ROM'...");
				builder.createFile(PC88FontBuilder::ROM_HIRAFONT);
				printf("done.\n");
			}
			printf("completed.\n");
		}
	} catch(const char* error) {
		printf("\nError : %s.\n", error);
	}
	return 0;
}
